/*
 * Programming in Java, Jacobs University, 2015
 *
 * This class demonstrates the usage of the Callable interface and of
 * Futures, an abstraction to organize concurrent computations which
 * is gaining in popularity. This version shows the usage of locks to
 * control access to a shared resource.
 *
 * $ java E a b d e f gg ff gg a abc
 */

import java.lang.Thread;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

public class E implements Callable<String>
{
    private static volatile String r = "";
    private static final ReentrantReadWriteLock rwl
	= new ReentrantReadWriteLock();
    private final String s;
    private int ms = 0;

    public E(String s)
    {
	this(s, 1000);
    }

    public E(String s, int ms)
    {
	this.s = s;
	this.ms = ms;
    }

    public String call() throws Exception
    {
	int len = s.length();
	String result = null;
	
	for (int i = 0; i < len && ! Thread.currentThread().isInterrupted(); i++) {
	    rwl.writeLock().lock();
	    try {
		r += s + " ";
		result = r;
	    } finally {
		rwl.writeLock().unlock();
	    }
	    try {
		TimeUnit.MILLISECONDS.sleep(ms);
	    } catch(InterruptedException e) {
		Thread.currentThread().interrupt();
	    }
	}
	return result;
    }

    public static void main(String[] args)
    	throws InterruptedException, ExecutionException
    {
	ExecutorService pool = Executors.newFixedThreadPool(3);
	List<Future<String>> futures = new ArrayList<Future<String>>();

	for (String arg: args) {
	    futures.add(pool.submit(new E(arg, 42)));
	}

	for (Future<String> f: futures) {
	    System.out.println(f.get());
	}

        pool.shutdown();
    }
}
